"""
Plotting networks
"""
import numpy as np
import networkx as nx
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle, FancyArrowPatch, Arc, Polygon
from matplotlib.patches import ArrowStyle

# Edge colors
activ = plt.get_cmap('tab10')(2)
inhib = plt.get_cmap('tab10')(3)

def build_pos(inter, method=None):
    """
    Compute node layout given an interaction matrix.
    """
    G = inter.shape[0]
    # Define graph
    graph = nx.Graph()
    for i in range(G):
        graph.add_node(i)
        for j in range(0,G):
            graph.add_node(j)
            if np.abs(inter[i,j] != 0):
                graph.add_edge(i, j)
    # Compute graph layout
    if method is None:
        p0 = np.random.normal(size=(G,2))
        p = nx.kamada_kawai_layout(graph, pos=p0)
    if method == 'graphviz':
        p = nx.nx_agraph.graphviz_layout(graph, prog='neato')
    # Return node positions
    pos = np.zeros((G,2))
    for k in range(0,G):
        pos[k,:] = p[k]
    scale = 1 if method is None else 2/(pos.max()-pos.min())
    center = pos.mean(axis=0)
    return scale * (pos - center)

#### Plotting functions ####

def node(k, ax, pos, name, scale=1, color=None, fontsize=None, nodesize=1):
    if color is None:
        color = 'gray'
    if fontsize is None:
        fontsize = 8*scale
    x, y = pos[k,0] * scale, pos[k,1] * scale
    r = nodesize * 0.1 * scale
    # Node shape
    circle1 = plt.Circle((x, y), radius=r, fill=True, facecolor='white',
        edgecolor='lightgray', lw=1*scale, zorder=3, clip_on=False)
    ax.add_artist(circle1)
    # Node label
    ax.text(x, y - 0.007*scale, name, color=color, fontsize=fontsize,
        zorder=4, horizontalalignment='center', verticalalignment='center')

def link(k1, k2, ax, pos, weight, bend=0, scale=1, nodesize=1, alpha=None):
    # Node coordinates
    x1, y1 = pos[k1,0]*scale, pos[k1,1]*scale
    x2, y2 = pos[k2,0]*scale, pos[k2,1]*scale
    shrink = 9 * scale * nodesize
    if alpha is None: alpha = 1

    # Case 1: activation
    if weight > 0:
        style = ArrowStyle('Simple', tail_width=1.1*scale,
            head_width=3.5*scale, head_length=5*scale)
        arrow = FancyArrowPatch((x1,y1), (x2,y2), arrowstyle=style,
        shrinkA=shrink, shrinkB=shrink, fc=activ, lw=0, zorder=0,
        connectionstyle='arc3,rad={}'.format(bend),
        clip_on=False, alpha=alpha)
        ax.add_artist(arrow)

    # Case 2: inhibition
    if weight < 0:
        style = ArrowStyle('Simple', tail_width=1.1*scale,
            head_width=0, head_length=1e-9*scale)
        arrow = FancyArrowPatch((x1,y1), (x2,y2), arrowstyle=style,
        shrinkA=shrink, shrinkB=shrink, fc=inhib, lw=0, zorder=0,
        connectionstyle='arc3,rad={}'.format(bend),
        clip_on=False, alpha=alpha)
        ax.add_artist(arrow)

        r = 0.125*nodesize
        u = pos[k2] - pos[k1]
        u0 = u/np.sqrt(np.sum(u**2))
        x0 = (pos[k1,0] + r*u0[0])*scale
        y0 = (pos[k1,1] + r*u0[1])*scale
        dx = (u[0] - 2*r*u0[0])*scale
        dy = (u[1] - 2*r*u0[1])*scale

        h_width = 0.07*scale
        h_height = 0.015*scale
        x1, y1 = x0 + dx, y0 + dy
        d = np.sqrt(dx**2 + dy**2)
        u = np.array([dx, dy])/d
        v = np.array([dy,-dx])/d

        x = x1 + 0.2*h_height*u[0] + 0.5*h_width*v[0]
        y = y1 + 0.2*h_height*u[1] + 0.5*h_width*v[1]

        if v[0] > 0: angle = (180/np.pi)*np.arctan(v[1]/v[0]) - 180
        elif v[0] < 0: angle = (180/np.pi)*np.arctan(v[1]/v[0])
        else: angle = np.sign(u[0]) * 90

        angle += np.tanh(bend) * 90
        theta = np.tanh(bend) * (np.pi/2)

        x3 = x2 + np.cos(theta)*(x-x2) - np.sin(theta)*(y-y2)
        y3 = y2 + np.sin(theta)*(x-x2) + np.cos(theta)*(y-y2)

        head = Rectangle((x3,y3), h_width, h_height, clip_on=False,
            angle=angle, fc=inhib, zorder=0, alpha=alpha)
        ax.add_artist(head)

def link_auto(k, ax, pos, weight, v=None, scale=1, nodesize=1, alpha=None):
    # Node coordinates
    x0, y0 = pos[k,0]*scale, pos[k,1]*scale

    # Orientation
    if v is None:
        v = np.array([1,0])
        v /= np.sqrt(np.sum(v**2))

    if v[0] > 0: angle0 = np.arcsin(v[1]) * (180/np.pi)
    else: angle0 = 180 - np.arcsin(v[1]) * (180/np.pi)

    angle1 = 58 + angle0
    angle2 = 300 + angle0
    theta = angle0 * (np.pi/180)

    x, y = x0 + 0.145*scale, y0
    x1 = x0 + np.cos(theta)*(x-x0) - np.sin(theta)*(y-y0)
    y1 = y0 + np.sin(theta)*(x-x0) + np.cos(theta)*(y-y0)

    # Loop diameter
    d = 0.15*scale

    # Case 1: activation
    if weight > 0:
        c = activ
        arc = Arc((x1,y1), d, d, angle=180, theta1=angle1, theta2=angle2-30,
        lw=1.1*scale, color=c, zorder=0, clip_on=False)
        ax.add_artist(arc)

        # Activation head
        head_width = 0.045*scale
        head_length = 0.065*scale

        angle = angle2 - 180 - 30
        theta = angle * (np.pi/180)

        triangle = np.array([
            [x1 + d/2 + head_width/2, y1 - head_length/4],
            [x1 + d/2, y1 + head_length * 3/4],
            [x1 + d/2 - head_width/2, y1 - head_length/4]])
        x, y = triangle[:,0].copy(), triangle[:,1].copy()
        triangle[:,0] = x1 + np.cos(theta)*(x-x1) - np.sin(theta)*(y-y1)
        triangle[:,1] = y1 + np.sin(theta)*(x-x1) + np.cos(theta)*(y-y1)
        triangle -= 0.03 * (triangle[0]-triangle[2])

        head = Polygon(triangle, fc=activ, zorder=0, clip_on=False)
        ax.add_artist(head)

    # Case 2: inhibition
    if weight < 0:
        c = inhib
        arc = Arc((x1,y1), d, d, angle=180, theta1=angle1, theta2=angle2-2,
        lw=1.1*scale, color=c, zorder=0, clip_on=False)
        ax.add_artist(arc)

        # Inhibition head
        width = 0.06*scale
        height = 0.015*scale

        angle = angle2 - 180 - 2
        theta = angle * (np.pi/180)

        x = x1 - width/2 + d/2
        y = y1 - height/2

        x2 = x1 + np.cos(theta)*(x-x1) - np.sin(theta)*(y-y1)
        y2 = y1 + np.sin(theta)*(x-x1) + np.cos(theta)*(y-y1)

        head = Rectangle((x2,y2), width, height, angle=angle, fc=c,
            zorder=0, clip_on=False)
        ax.add_artist(head)

#### Main function ####

def plot_network(inter, pos, width=1, height=1, scale=1, names=None,
    vdict=None, tol=None, root=False, axes=None, nodes=None, n0=True,
    file=None, verb=False, fontsize=None, vcolor=None, nodesize=1,
    bend=0.14, bend_all=False, alpha=None):
    """
    Plot a network.
    """
    G, G = inter.shape
    w, h = width/2.54, height/2.54 # Centimeters
    if names is None: names = [''] + ['{}'.format(k) for k in range(1,G)]
    if vcolor is None: vcolor = ['#5C5C5C' for k in range(G)]
    if vdict is None: vdict = {}
    # Compute layout
    v = list(range(G))
    e = list(zip(*inter.nonzero()))
    # Draw layout
    if axes is None:
        fig = plt.figure(figsize=(w,h), dpi=100, frameon=False)
        plt.axes([0,0,1,1])
        ax = fig.gca()
        I, J = inter.nonzero()
        ax.axis('off')
    else:
        ax = axes
        fig = plt.gcf()
        size = fig.get_size_inches()
        w, h = size[0], size[1]
        if nodes is None: I, J = inter.nonzero()
        else: I, J = nodes, nodes
    ax.axis('equal')
    ax.axis('off')
    ax_pos = ax.get_position()
    plt.xlim([-w*ax_pos.width/2, w*ax_pos.width/2])
    plt.ylim([-h*ax_pos.height/2, h*ax_pos.height/2])
    scale = scale * np.min([ax_pos.width,ax_pos.height])
    # Draw nodes
    for k in range(G):
        node(k, ax, pos, names[k], scale, fontsize=fontsize, color=vcolor[k],
            nodesize=nodesize)
    # Draw links
    for k1, k2 in e:
        weight = inter[k1,k2]
        if k1 != k2:
            if (k2, k1) in e:
                link(k1, k2, ax, pos, weight, bend, scale, nodesize=nodesize,
                    alpha=alpha)
            else: link(k1, k2, ax, pos, weight, bend*bend_all, scale,
                nodesize=nodesize, alpha=alpha)
        else:
            v = vdict.get(k1)
            if v is None:
                b, c = 0, 0
                for k in range(G):
                    if (k != k1) and (((k,k1) in e) or ((k1, k) in e)):
                        b += pos[k]
                        c += 1
                if c == 0:
                    b, c = 0, 0
                    for k in I | J:
                        if (k != k1):
                            b += pos[k]
                            c += 1
                v = pos[k1] - b/c
            else: v = np.array(v)
            d = np.sqrt(np.sum(v**2))
            if d > 0: v = v/d
            else: v = np.array([0,1])
            link_auto(k1, ax, pos, weight, v, scale, nodesize=nodesize,
                alpha=alpha)
    if file is None: file = 'network.pdf'
    if axes is None: fig.savefig(file, bbox_inches='tight')



"""
harissa.utils
-------------

Various utility functions.
"""
from .plot_network import build_pos, plot_network
from .processing import binarize

__all__ = ['build_pos', 'plot_network', 'binarize']



"""
Harissa
=======

Tools for mechanistic gene network inference from single-cell data
------------------------------------------------------------------

Mechanistic model-based gene network inference using
a self-consistent proteomic field (SCPF) approximation.
It is analogous to the unrestricted Hartree approximation
in quantum mechanics, applied to gene expression modeled
as a piecewise-deterministic Markov process (PDMP).

The package also includes a simulation module to generate
single-cell data with transcriptional bursting.

Author: Ulysse Herbach (ulysse.herbach@inria.fr)
"""
__version__ = '3.0.7'

from .model import NetworkModel, Cascade, Tree

__all__ = ['NetworkModel', 'Cascade', 'Tree']



"""
Perform simulations using the PDMP model - Fast version using Numba
NB: This module takes time to compile (~8s) but is much more efficient,
which is typically suited for large numbers of genes and/or cells
"""
import numpy as np
from numba import njit

class BurstyPDMP:
    """
    Bursty PDMP version of the network model (promoters not described)
    """
    def __init__(self, a, d, basal, inter, thin_adapt=True):
        # Kinetic parameters
        G = basal.size
        D0, D1 = d[0], d[1]
        K0, K1, B = a[0]*d[0], a[1]*d[0], a[2]
        S1 = D0*D1*a[2]/K1 # Normalize protein scales
        types = [('D0','float'), ('D1','float'), ('S1','float'),
            ('K0','float'), ('K1','float'), ('B','float')]
        plist = [(D0[i], D1[i], S1[i], K0[i], K1[i], B[i]) for i in range(G)]
        self.param = np.array(plist, dtype=types)
        # Network parameters
        self.basal = basal
        self.inter = inter
        # Default state
        types = [('M','float'), ('P','float')]
        self.state = np.array([(0,0) for i in range(G)], dtype=types)
        # Thinning parameter
        self.thin_cst = None if thin_adapt else np.sum(K1[1:])

    def simulation(self, timepoints, verb=False):
        """
        Exact simulation of the network in the bursty PDMP case.
        """
        basal, inter = self.basal, self.inter
        K0, K1, B = self.param['K0'], self.param['K1'], self.param['B']
        D0, D1, S1 = self.param['D0'], self.param['D1'], self.param['S1']
        M, P, thin = self.state['M'], self.state['P'].copy(), self.thin_cst
        if np.size(timepoints) == 1: timepoints = np.array([timepoints])
        timepoints = timepoints.astype(float)
        # Compute the simulation
        states = simulate(timepoints, basal, inter,
            K0, K1, B, D0, D1, S1, M, P, thin, verb)
        # Update the current state
        self.state['M'], self.state['P'] = states[0,-1], states[1,-1]
        # Store the results
        types = [('M','float64'), ('P','float64')]
        G, sim = basal.size, []
        for k, t in enumerate(timepoints):
            M, P = states[0,k], states[1,k]
            sim += [np.array([(M[i],P[i]) for i in range(1,G)], dtype=types)]
        return np.array(sim)

# Core functions for Numba acceleration

@njit
def kon(basal, inter, K0, K1, P):
    """
    Interaction function kon (off->on rate), given protein levels p.
    """
    Phi = np.exp(basal + P @ inter)
    Kon = (K0 + K1*Phi)/(1 + Phi)
    Kon[0] = 0 # Ignore stimulus
    return Kon

@njit
def kon_bound(basal, inter, K0, K1, D0, D1, S1, M, P):
    """
    Compute the current kon upper bound.
    """
    # Explicit upper bound for P
    time = np.log(D0/D1)/(D0-D1) # Vector of critical times
    pmax = P + (S1/(D0-D1))*M*(np.exp(-time*D1) - np.exp(-time*D0))
    pmax[0] = P[0] # Discard stimulus
    # Explicit upper bound for Kon
    Phi = np.exp(basal + pmax @ ((inter > 0) * inter))
    Kon = (K0 + K1*Phi)/(1 + Phi) + 1e-10 # Fix precision errors
    Kon[0] = 0 # Ignore stimulus
    return Kon

@njit
def flow(D0, D1, S1, M, P, time):
    """
    Deterministic flow for the bursty model.
    """
    Mnew = M*np.exp(-time*D0)
    Pnew = ((S1/(D0-D1))*M*(np.exp(-time*D1) - np.exp(-time*D0))
            + P*np.exp(-time*D1))
    Mnew[0], Pnew[0] = M[0], P[0] # Discard stimulus
    state = np.zeros((2,M.size))
    state[0], state[1] = Mnew, Pnew
    return state

@njit
def step(basal, inter, K0, K1, B, D0, D1, S1, M, P, thin):
    """
    Compute the next jump and the next step of the
    thinning method, in the case of the bursty model.
    """
    if thin is None:
        # Adaptive thinning parameter
        tau = np.sum(kon_bound(basal, inter, K0, K1, D0, D1, S1, M, P))
    else: tau = thin
    jump = False # Test if the jump is a true or phantom jump
    # 0. Draw the waiting time before the next jump
    U = np.random.exponential(scale=1/tau)
    # 1. Update the continuous states
    state = flow(D0, D1, S1, M, P, U)
    M[:], P[:] = state[0], state[1]
    # 2. Compute the next jump
    v = kon(basal, inter, K0, K1, P)/tau # i = 1, ..., G-1 : burst of mRNA i
    v[0] = 1 - np.sum(v[1:]) # i = 0 : no change (phantom jump)
    i = np.nonzero(np.random.multinomial(1, v))[0][0]
    if i > 0:
        r = B[i]
        M[i] += np.random.exponential(1/r)
        jump = True
    return U, jump

@njit
def simulate(timepoints, basal, inter, K0, K1, B, D0, D1, S1, M, P,
    thin, verb):
    """
    Exact simulation of the network in the bursty PDMP case.
    """
    states = np.zeros((2,timepoints.size,basal.size))
    # sim = np.zeros(len(timepoints), dtype=types)
    c0, c1 = 0, 0 # Jump counts (phantom and true)
    T = 0
    # Core loop for simulation and recording
    Told, Mold, Pold = T, M.copy(), P.copy()
    for k, t in enumerate(timepoints):
        while T < t:
            Told, Mold, Pold = T, M.copy(), P.copy()
            U, jump = step(basal, inter, K0, K1, B, D0, D1, S1, M, P, thin)
            T += U
            if jump: c1 += 1
            else: c0 += 1
        state = flow(D0, D1, S1, Mold, Pold, t - Told)
        states[0,k], states[1,k] = state[0], state[1]
    # Display info about jumps
    if verb:
        ctot = c0 + c1
        if ctot > 0:
            print(f'Exact simulation used {ctot} jumps '
                + f'including {c0} phantom jumps '
                + f'({int(100*c0/ctot)}%)')
        else: print('Exact simulation used no jump')
    return states



from setuptools import setup, find_packages
import harissa

setup(

    name='harissa',

    version=harissa.__version__,

    packages=find_packages(),

    author='Ulysse Herbach',

    author_email='ulysse.herbach@inria.fr',

    description=('Tools for mechanistic gene network inference '
        'from single-cell data'),

    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',

    url='https://github.com/ulysseherbach/harissa',

    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ],

    platforms='any',

    license='BSD-3-Clause',
    license_file='LICENSE.txt',

    keywords=('stochastic gene expression, gene regulatory networks, '
        'single-cell transcriptomics'),

    python_requires='>=3.8',

    install_requires=[
        'numpy>=1.20',
        'scipy>=1.7',
        'numba>=0.55',
        'matplotlib>=3.4',
        'networkx>=2.6',
    ],

)
